/*
 * Routines for filtering the data.
 * filter_data is the top-level routine, which handles all
 * the steps necessary.
 */

#include <stdio.h>
#include <limits.h>
#include <math.h>
#include <stdlib.h>
#include <limits.h>

#include "filter.h"


static int NINT(double t);

static void freq_filter(double filt[], int n2, float dinterv, float cutoff);
static void lin_interp(
	double x[],
	double y[],
	int n,
	double x1[],
	double yp[],
	int np,
	float dinterv
);

static void adjustend (
	double x[],
	double y[],
	int n,
	int cutoff,
	float *a,
	float *b
);


#define TWOPI 		6.283185307
#define MAX_PARAM	30

static int numpt, numhm, numpm;

/************************************************************************/
void filter_data (

	double 	x[],		/* Input - array containing sample dates */
	double 	y[],		/* Input - array containing measurement values */
	double 	sig[],		/* Input - array containing uncertainties of measurements */
	int 	n0,		/* Input - number of points in x, y, sig */
	int 	cutoff1,	/* Input - short term cutoff value in days */
	int 	cutoff2,	/* Input - long term cutoff value in days */
	float 	interval,	/* Input - regularly spaced sample interval in days */
	int 	numpoly,	/* Input - number of polynomial terms in function fit */
	int 	numharm,	/* Input - number of harmonic terms in function fit */
	float 	pm[],		/* Output - array containing coefficients of function fit */
	double 	**covar,	/* Output - array containing covariance matrix */
	int 	*numparam,	/* Output - Total number of parameters in pm */
	double 	xfilt[],	/* Output - dates of filtered data points */
	double 	smooth[],	/* Output - measurement values of filtered data */
	double 	trend[],	/* Output - Filter results of residuals using long term cutoff */
	double 	deriv[],	/* Output - derivative of trend + polynomial, i.e. growth rate */
	int 	*nfilt,		/* Output - number of points in xfilt, smooth, trend, deriv */
	float 	*varf1,		/* Output - variance of filter with short term cutoff */
	float 	*varf2,		/* Output - variance of filter with long term cutoff */
	float 	*chisq,		/* Output - chi squared value of function fit */
	float 	*rsd1		/* Output - residual standard deviation of points about the function */
)

{
	float rinterval, cutoff, rmean, ca, cb;
	double dp, p;
	int i, j, np;
	double *work1, *work2, *yp;

	numpt = numpoly;
	numhm = numharm;
	numpm = numpoly + 2*numharm;
	*numparam = numpm;

	work1 = (double *) calloc (n0, sizeof(double));



#ifdef DEBUG
printf ("Inside filter_data.  n0 = %d Interval = %f\n", n0, interval);
printf ("Cutoff 1 = %d, Cutoff 2 = %d\n",cutoff1,cutoff2);
printf ("Numpoly = %d, Numharm = %d\n", numpoly,numharm);
printf ("First point = %e,%e\n", x[0], y[0]);
printf ("Last point = %e,%e\n",x[n0-1], y[n0-1]);
#endif

	/*
	 * Fit the function to the data
	 */
	if (numpm > 0) {
		lfit (x, y, sig, n0, pm, covar, numpm, chisq, partial);
	}

#ifdef DEBUG
printf ("Finished lfit\n");
for (i=0; i< numpm; i++) {
  printf ("pm[%d] = %e\n", i, pm[i]);
}
#endif

	/*
	 *  calculate residuals from fit
	 */
	for (i=0; i<n0; i++) {
		work1[i] = y[i] - userfunc (pm, x[i], numhm, numpt);
	}
	means (work1, &rmean, rsd1, n0);

#ifdef DEBUG
printf ( "Finished residuals, rmean = %e, rsd = %e\n", rmean, *rsd1);
#endif

	/*
	 *  Remove a linear trend based on the end points of the residuals.
	 *  This is to reduce the end effects of the filter.
	 */

	adjustend (x, work1, n0, cutoff2, &ca, &cb);

#ifdef DEBUG
printf ("Finished adjustend, ca = %e, cb = %e\n", ca,cb);
printf ("x[0] = %e, x[n0-1] = %e\n", x[0], x[n0-1]);
printf ("work1[0] = %e, work1[n0-1] = %e\n", work1[0], work1[n0-1]);
#endif

	/* 
	 * Determine how many points there will be from first data point
	 * to last data point using equally spaced intervals.
	 */

	rinterval = interval / 365.0;
	dp = (x[n0-1] - x[0])/rinterval;
	np = NINT(dp) + 1;
	*nfilt = np;

#ifdef DEBUG
printf ("np = %d\n", np);
#endif

	/* Allocate an array to hold interpolated data */
	/* Then interpolate to get equally spaced data at sampling interval */
	yp = (double *) calloc(np, sizeof(double));
	lin_interp(x, work1, n0, xfilt, yp, np, rinterval);

#ifdef DEBUG
printf ("Interpolated points.\n");
printf ("xfilt[np-1] = %e, x[0] = %e\n", xfilt[np-1], x[0]);
printf ("yp[np-1] = %e, y[0] = %e\n", yp[np-1], yp[0]);
printf ("work1[0] = %e\n", work1[0]);
#endif


	/* Filter the residuals with the short term cutoff value */
	cutoff = (float) cutoff1 / 365.0;
	work2 = filter (yp, np, rinterval, cutoff);

#ifdef DEBUG
printf ("Finished short term filter. nfilt = %d\n", *nfilt);
#endif


	/*
	 * Add back the linear fit to the ends of the data.
	 */
	for (i=0; i< *nfilt; i++) smooth[i] = work2[i] + ca + cb*xfilt[i];

	/*  Find variance of short term filter */
	*varf1 = filtvar (cutoff, rinterval, xfilt, smooth, np, x, work1, n0);

#ifdef DEBUG
printf ("Finished short term filter variance = %e\n", *varf1);
#endif

	/* Filter the residuals with the long term cutoff value */
	cutoff = (float) cutoff2/365.0;
	work2 = filter (yp, np, rinterval, cutoff);

#ifdef DEBUG
printf ("Finished long term filter\n");
#endif

	for (i=0; i< *nfilt; i++) trend[i] = work2[i] + ca + cb*xfilt[i];

	/*  Find variance of long term filter */
	*varf2 = filtvar(cutoff, rinterval,xfilt, trend, np, x, work1, n0);

#ifdef DEBUG
printf ("Finished long term filter variance = %e\n", *varf2);
#endif

	/* compute derivative of trend */
	spline_f (xfilt, trend, np, xfilt, work2, deriv, np);

#ifdef DEBUG
printf ("Finished spline\n");
#endif

	/* Find derivative of polynomial at each filtered point, add to trend derivative */
	if (numpoly > 0) {
		for (i=0; i< np; i++) {
			dp = 0.0;
			p = pm[numpoly-1];
			for (j=numpoly-2; j>=0; j--) {
				dp = dp*xfilt[i] + p;
				p = p*xfilt[i] + pm[j];
			}
			deriv[i] += dp;
		}
	}

	free (work1);
	free (work2);
	free (yp);

#ifdef DEBUG
printf ("Finished derivative.  Returning ...\n");
#endif

      return;
}

/************************************************************************/
/*
 * Given a time series specified by x[] and y[] of lenth n, create
 * another series at times domain[] of length narg with interpolated values 
 * in func[] and the first derivative values in deriv[].
 */
/************************************************************************/
void spline_f (

	double 	x[],		/* Input - array containing x values */
	double 	y[],		/* Input - array containing y values */
	int 	n,		/* Input - Number of points in x, y */
	double 	domain[],	/* Input - where to compute interpolated values along x axis */
	double 	func[],		/* Output - the interpolated values */
	double 	deriv[],	/* Output - the derivative of the interpolated values */
	int 	narg		/* Input - the number of points in domain, func, deriv */
)

{
	double *y2;
	int i;

#ifdef DEBUG
printf ("inside spline, n = %d, narg = %d\n", n, narg);
#endif

	y2 = (double *) calloc (n, sizeof(double));

	spline (x, y, n, y2);

	for (i=0; i<narg; i++) {
		splint (x, y, y2, n, domain[i], &func[i], &deriv[i]);
	}

	free (y2);
}


/************************************************************************/
/*  User supplied function defining curve fit.
 *  Function is a combination of polynomial terms and
 *  sine, cosine harmonic terms.
 *  f = a1*x + a2*x^2 + X3*x^3 ... for a = 1 to nt  +
 *      p[2*i-1]*sin(2pi*x) + p[2*i]*cos(2pi*x) + ... for i = 1 to nh
 */
/************************************************************************/
double userfunc (

	float 	pm[],	/* Input - array of parameters from function fit */
	double 	x,	/* Input - value at which to calculate the function */
	int 	nh,	/* input - number of harmonic terms in p[] */
	int	nt	/* Input - number of polynomial terns in p[] */
)

{
	double v, p;

	v = yearharmonic (pm, x, nt, nh);
	p = polyv(pm, x, nt);

	return (v+p);
}

/************************************************************************/
/*
 * Evaluate the polynomial part of the function at time x.
 */
/************************************************************************/
double polyv (

	float   pm[],	/* Input - array of parameters from function fit */
	double	x,	/* Input - Value at which to calculate polynomial */
	int 	numpt	/* Input - Number of polynomial terms in pm[] */
)

{
	double p;
	int j;

	if (numpt <= 0) {
		return (0.0);
	}

	p = pm[numpt-1];
	for (j=numpt-2; j >= 0; j--) {
		p = p*x + pm[j];
	}

	return (p);
}

/************************************************************************/
/*
 * Evaluate the harmonic part of the function at time x.
 */
/************************************************************************/
double yearharmonic (

	float pm[],	/* Input - array of parameters from function fit */
	double	x,	/* Input - Value at which to calculate polynomial */
	int 	numpt,	/* Input - Number of polynomial terms in pm[] */
	int 	numh	/* Input - Number of harmonic terms in pm[] */
)

{
	double s, twopix;
	int i, ix;

	s = 0.0;
	twopix = TWOPI*x;
	for (i=0; i<numh; i++) {
		ix = 2*i + numpt;
		s += pm[ix]*sin((i+1)*twopix) + pm[ix+1]*cos(twopix*(i+1));
	}

	return (s);
}

/************************************************************************/
/*
 * Calculate the mean and standard deviation from array.
 */
/************************************************************************/
void means (

	double 	array[],/* Input - array of values to calculate mean */
	float  	*rmean,	/* Output - mean value */
	float	*sdev,	/* Output - standard deviation */
	int	n	/* Input - number of values in a[] */
)

{
	double sum, sumsq;
	int i;
	
	*rmean = 0.0;
	*sdev = 0.0;
	if (n < 1) {
		fprintf (stderr, "Warning: n < 1 in routine means.\n");
		return;
	}

	sum = 0.0;
	sumsq = 0.0;
	for (i=0; i<n; i++) {
		sum += array[i];
		sumsq += (array[i]*array[i]);
	}

	*rmean = sum/((float) n);
	if (n > 1) {
		*sdev = sqrt ((sumsq - (sum*sum)/(float) n) / ((float) (n-1)));
	}
	return;
}


/************************************************************************/
/*
 * Given time series x[] and y[] of length n, find the points that are
 * within cutoff/4 of the ends, then compute linear regression on those
 * points with intercept a and slope b.
 */
/************************************************************************/
static void adjustend (

	double x[],	/* Input - array of time values */
	double y[],	/* Input - array of values */
	int n,		/* Input - length of x[] and y[] */
	int cutoff,	/* Input - cutoff value of filter */
	float *a,	/* Output - intercept of linear regression line */
	float *b	/* Output - slope of linear regression line */
)

{
	double *xp, *yp;
	int k, i;
	float c, siga, sigb, chi2;
	

	xp = (double *) calloc (n, sizeof(double));
	yp = (double *) calloc (n, sizeof(double));

	*a = 0.0;
	*b = 0.0;
	if ((x[n-1] - x[0]) < (float) (cutoff) / 365.0) return;

	c = (float) cutoff /365.0 / 4.0;
	k = 0;
	for (i=0; i<n; i++) {
		if ((x[i] <= x[0]+c) || (x[i] >= x[n-1]-c)) {
			xp[k] = x[i];
			yp[k] = y[i];
			k++;
		}
	}

	fit (xp, yp, k, a, b, &siga, &sigb, &chi2);

	for (i=0; i<n; i++) {
		y[i] -= (*a + *b*x[i]);
	}

	free(xp);
	free(yp);

	return;
}


/*********************************************************************/
/*
 * Calculate the partial derivatives of user function
 * with respect to each parameter
 *
 * x   - Value at which to find partial derivative
 * der - Array of partial derivatives at time x
 * numpm - Number of parameters in der
 *
 * Calls routine 'part(i,n)' which returns the value of the 
 * partial derivative with respect to parameter i at time x.
 * Written this way since the routine varnce needs to use 'part' also.
 *
 */
/*********************************************************************/
void partial (

	double x,
	double der[],
	int numpm
)
{
	int i;

	for (i=0; i<numpm; i++) {
		der[i] = part(i,x);
	}
}

/*********************************************************************/
/*
 * Find partial derivative of user function for parameter n at x
 * with respect to coefficients.
 *
 * n - Number of the parameter 0..numpm-1
 * x - Value at which to calculate the partial derivative
 * numpt - number of polynomial terms in user function
 *
 */
/*********************************************************************/
double part(
	int n,
	double x
)
{
	double p = 0.0, xx;
	int ix;

	if (n < numpt) {
		if ( n==0) {
			p = 1.0;
		} else {
			p = pow (x, (double) n);
		}
	} else {
		ix = ((n-numpt) /2 + 1);
		xx = ix * TWOPI * x;
		if ((n-numpt) % 2 == 0) {
			p = sin(xx);
		} else {
			p = cos(xx);
		}
	}
	return (p);
}

/*****************************************************/
/* 
 * Perform a fast fourier transform on the data, 
 * then apply a low pass filter funtion to the frequency data,
 * then inverse transform back to time domain.
 *
 * The realft subroutine from Press et al requires that the
 * length of the data array be 2^n (256, 512, 1024...).  So
 * zero pad the data on each end to get to the correct amount
 * of points.
 *
 * Returns an array of filtered data same length as yp
 */
/*****************************************************/
double *filter (
	double yp[],	/* Input - array of values */
	int np,		/* Input - length on yp */
	float dinterv,  /* Input - interval between data points */
	float cutoff	/* Input - filter cutoff value */
)
{
	int n2, i, numzer, n3;
	double *filt;

	/* Determine the number of points >= np that is a power of 2 */

        n2 = 1;
        while (np > n2) {
                n2 = n2<<1;
        }

        if (n2 > MAX_OBS) {
                fprintf (stderr, "Error:  Too many data points to filter.\n");
                fprintf (stderr, "Maybe try using larger interval size.\n");
                exit(1);
        }

	filt = (double *) calloc(n2+2, sizeof(double));
        numzer = n2 - np;
        n3 = numzer/2;	/* this is where the real data will begin in the zero padded array */
	for (i=n3; i<n3+np; i++) {
		filt[i] = yp[i-n3];
	}

	/* Transform the data (time >> frequency) */
        realft (filt-1, n2, 1);

#ifdef DEBUG
printf ("Finished forward fft\n");
#endif

        freq_filter(filt, n2, dinterv, cutoff);
/*
        for ( i = 0; i < nc; i++ ) {
                printf ( "  %4d  %12f  %12f\n", i, out[i][0], out[i][1] );
        }
exit(0);
*/


#ifdef DEBUG
printf ("Finished lowpass filter response\n");
#endif

	/* Inverse transform the data (frequency >> time) */
        realft (filt-1, n2, -1);

#ifdef DEBUG
printf ("Finished inverse fft\n");
#endif

	/* remove zero padding */
        for (i=0; i< np; i++) {
                filt[i] = filt[n3+i];
        }


	return filt;
}

/***********************************************************/
/* compute equally spaced values using linear interpolation */
/***********************************************************/
static void lin_interp(
	double x[],	/* Input - Array of time values */
	double y[],	/* Input - Array of measurement values */
	int n,		/* Input - Length of x and y */
	double x1[],	/* Output - Interpolated time values */
	double yp[],	/* Output - Interpolated measurement values */
	int np,		/* Input - Length of x1, yp */
	float dinterv	/* Input - Time interval between interpolated data points */
)
{
	int i, j;
	double h, w;


	i=0;
	x1[0] = x[0];
	yp[0] = y[0];
	for (j=1; j< np; j++) {
		x1[j] = x[0] + j*dinterv;
		/* Find x value greater than x1[j] */
		/* Complicated by the fact that the
		   x values could be pairs, meaning the last
		   2 points have the same x value */
		while (x1[j] >= x[i]) {
			i++;
			if (i >= (n-1)) break;
                }

		i--;

		/* If we are past the last actual data point, 
		   and the last two points were a pair, then
		   the x data points will be the same,
		   making w infinite. So instead of extrapolating,
		   just set the point to 0 (which means assume
		   it is equal to the function value at that time). */
		if (x1[j] > x[n-1]) { 
			yp[j] = 0.0; 
		} else {
			h = x1[j] - x[i];
			w = (y[i+1]-y[i])/(x[i+1]-x[i]);
			yp[j] = w*h + y[i];
		}

	}

}


/***************************************************************/
/*
 * filter the data by multiplying the frequency domain data
 * with the response of a low pass filter.
 * b - frequency step size
 * dmaxfreq - nyquist frequency
 */
/***************************************************************/
static void freq_filter(
	double filt[], 
	int n2, 
	float dinterv, 
	float cutoff
)
{
        float b, dmaxfreq, freq, cutoff2, rw;
        int i;

        b = 1.0/(n2*dinterv);
        dmaxfreq = 1.0/(2*dinterv);
        cutoff2 = 1.0/cutoff;
        for (i=2; i<n2+2; i+=2) {
                freq = (float) i/2.0 * b;
                rw = fnfilt(freq, cutoff2, 6);
                filt[i] *= rw;
                filt[i+1] *= rw;
        }
	rw = fnfilt (dmaxfreq, cutoff2, 6);
	filt[1] *= rw;

}


/**********************************************************************/
float fnfilt (
	float freq, 
	float sigma, 
	int power
)
{
	float z, f;

	/* Watch out for underflow */

	z = pow ((double) freq/sigma, (double) power);
	if (z > 20.0) {
		f = 1e-10;
	} else {
		f = 1.0/(pow (2.0, (double) z));
	}
	return (f);
}
/*******************************************/
/* Find nearest integet to a double value. */
/*******************************************/
static int NINT(double t)
{
	if (t<0.0) {
		return ((int) t-.5);
	} else {
		if ((t+.5) > (double) (INT_MAX))
			return ((INT_MAX));
		else
			return ((int) (t+.5));
	}
}
